
// all the following functions deal with a triangle
// a, b and c denotes the edge length of the triangle;
// ra, rb and rc denotes the radius of the discs centered on the triangle vertices
// ri is the radius of the disc centered on the vertex opposite to the edge of length i

// angle opposite to side a
I angle(I a, I b, I c)
{
    return acos((b/c+c/b-square(a)/(b*c))/I(2));
}

// area of the triangle
I area(I a, I b, I c)
{
    return sqrt((a + b + c)*(a + b - c)*(a - b + c)*(- a + b + c))/I(4);
}

// area covered by disc sectors
I covered(I a, I b, I c, I ra, I rb, I rc)
{
    return (square(ra)*angle(a,b,c)+square(rb)*angle(b,c,a)+square(rc)*angle(c,a,b))/I(2);
}

// emptiness of the triangle
I E(I a, I b, I c, I ra, I rb, I rc)
{
    return delta*area(a,b,c)-covered(a,b,c,ra,rb,rc);
}

// radius of the support circle
I radius(I a, I b, I c, I ra, I rb, I rc)
{
    I S=(a+b+c)*(a+b-c)*(a-b+c)*(-a+b+c);
    I a2=square(a);
    I b2=square(b);
    I c2=square(c);
    I ra2=square(ra);
    I rb2=square(rb);
    I rc2=square(rc);
    I A=I(4)*((ra2-ra*rb-ra*rc+rb*rc)*a2 + (rb2-ra*rb+ra*rc-rb*rc)*b2 + (rc2+ra*rb-ra*rc-rb*rc)*c2)-S;
    I B=I(2)*(a2*(a2*ra-b2*(ra+rb)+(I(2)*ra2*ra-ra2*rb-ra*rb2-ra2*rc+rb2*rc-ra*rc2+rb*rc2)) + b2*(b2*rb-c2*(rb+rc)+(I(2)*rb2*rb-rb2*rc-rb*rc2-rb2*ra+rc2*ra-rb*ra2+rc*ra2)) + c2*(c2*rc-a2*(rc+ra)+(I(2)*rc2*rc-rc2*ra-rc*ra2-rc2*rb+ra2*rb-rc*rb2+ra*rb2)));
    I C=a2*b2*c2-(ra2+rb2)*a2*b2-(ra2+rc2)*a2*c2-(rb2+rc2)*b2*c2+(ra2*a2+square(ra2)-ra2*rb2-ra2*rc2+rb2*rc2)*a2+(rb2*b2-ra2*rb2+square(rb2)+ra2*rc2-rb2*rc2)*b2+(rc2*c2+ra2*rb2-ra2*rc2-rb2*rc2+square(rc2))*c2;
    I rs=I(-INFINITY,+INFINITY);
    // usual formula when A<>0
    if (not zero_in(A*C))
    {
        I D=I(4)*S*(a+rb-rc)*(a-rb+rc)*(b+ra-rc)*(b-ra+rc)*(c+ra-rb)*(c-ra+rb);
        rs=intersect(rs,(upper(C)<0 ? (-B+sqrt(D))/(I(2)*A) : (-B-sqrt(D))/(I(2)*A)));
    }
    // alternative formula (necessary if A contains 0, useful if A very small)
    I x=I(4)*A*C/square(B);
    x=hull(x,I(0)); // the interval x must contain 0
    if (upper(x)<0.78) rs=intersect(rs,-C/B*(I(1)+x));
    return rs;
}
        
// signed distance to edge a of the center of the support circle (positive iff the center is inside the triangle)
// rs is the radius of the support circle (to avoid computing it several times)
I dist(I a, I b, I c, I ra, I rb, I rc, I rs)
{
    I a2=square(a);
    I b2=square(b);
    I c2=square(c);
    I ra2=square(ra);
    I rb2=square(rb);
    I rc2=square(rc);
    I S=(a+b+c)*(a+b-c)*(a-b+c)*(-a+b+c);
    return ((b2+c2-a2 + I(2)*rs*(rb+rc-I(2)*ra)+rb2+rc2-I(2)*ra2)*a2 + (I(2)*rs+rb+rc)*(rb-rc)*(b2-c2))/(I(2)*a*sqrt(S));
}

// vertex potential
I Uv(I a, I b, I c, char ra, char rb, char rc)
{
    I Ra=val(ra);
    I Rb=val(rb);
    I Rc=val(rc);
    I u=min(I(ra=='1'?Z1:Zr)*(pi_twice<I>()-angle(a,b,c)),Vijk(ra,rb,rc)+I(ra=='1'?m1:mr)*abs(angle(a,b,c)-angle(Rb+Rc,Ra+Rc,Ra+Rb)));
    u+=min(I(rb=='1'?Z1:Zr)*(pi_twice<I>()-angle(b,c,a)),Vijk(rb,rc,ra)+I(rb=='1'?m1:mr)*abs(angle(b,c,a)-angle(Ra+Rc,Ra+Rb,Rb+Rc)));
    u+=min(I(rc=='1'?Z1:Zr)*(pi_twice<I>()-angle(c,a,b)),Vijk(rc,ra,rb)+I(rc=='1'?m1:mr)*abs(angle(c,a,b)-angle(Ra+Rb,Rb+Rc,Ra+Rc)));
    return u;
}

// parameters for the potential of edge ij
std::pair <double,double> LQ(char i, char j)
{
    if (i=='1' && j=='1') return std::pair<double,double>(l11,q11);
    else if (i!=j) return std::pair<double,double>(l1r,q1r);
    else return std::pair<double,double>(lrr,qrr);
}

// edge potential (rs=-1 if it has to be computed)
I Ue(I a, I b, I c, char ra, char rb, char rc, I rs=I(-1))
{
    I Ra=val(ra);
    I Rb=val(rb);
    I Rc=val(rc);
    if (rs<0) rs=radius(a,b,c,Ra,Rb,Rc); // compute the radius if not provided
    std::pair <double,double> lq=LQ(rb,rc);
    I u=I(0);
    if (upper(a)>lq.first) u+=lq.second*dist(a,b,c,Ra,Rb,Rc,rs);
    lq=LQ(ra,rc);
    if (upper(b)>lq.first) u+=lq.second*dist(b,c,a,Rb,Rc,Ra,rs);
    lq=LQ(ra,rb);
    if (upper(c)>lq.first) u+=lq.second*dist(c,a,b,Rc,Ra,Rb,rs);
    return u;
}

// total potential of the triangle (rs=-1 if it has to be computed)
I U(I a, I b, I c, char ra, char rb, char rc, I rs=I(-1))
{
    return Uv(a,b,c,ra,rb,rc)+Ue(a,b,c,ra,rb,rc,rs);
}

// <0 if the triangle can surely not appear in a saturated FM-triangulation, radius of support circle if it may appear
I feasible(I a, I b, I c, I Ra, I Rb, I Rc)
{
    // the altitude of each vertex must be at least r (no disc crosses the opposite edge)
    if (upper(square(b)-square((square(b)+square(c)-square(a))/(I(2)*c))-square(r))<0) return I(-1);
    if (upper(square(c)-square((square(c)+square(a)-square(b))/(I(2)*a))-square(r))<0) return I(-1);
    if (upper(square(a)-square((square(a)+square(b)-square(c))/(I(2)*b))-square(r))<0) return I(-1);
    // the area S must not be too small
    I S=(a+b+c)*(a+b-c)*(a-b+c)*(-a+b+c);
    if (upper(S)<lower(square(pi_twice<I>()*square(r)))) return I(-1);
    // the radius of support circle must be at most r
    I rs=radius(a,b,c,Ra,Rb,Rc);
    if (lower(rs)>upper(r)) return I(-1);
    return rs;
}

// recursive checking over smaller and smaller compact sets of triangles
int check_local_ieq(I a, I b, I c, char ra, char rb, char rc, int depth)
{
    I Ra=val(ra);
    I Rb=val(rb);
    I Rc=val(rc);    
    
    if (depth==25)
    {
        std::ostringstream err;
        err << "too deep!\n";
        I rs=feasible(a,b,c,Ra,Rb,Rc);
        I e=E(a,b,c,Ra,Rb,Rc);
        I u=U(a,b,c,ra,rb,rc,rs);
        I uv=Uv(a,b,c,ra,rb,rc);
        I ue=Ue(a,b,c,ra,rb,rc,rs);
        err << "E=[" << lower(e) <<", "<<upper(e)<<"], U=["<<lower(u)<<", "<<upper(u)<<"]\n";
        err << "Uv=[" << lower(uv) <<", "<<upper(uv)<<"], Ue=["<<lower(ue)<<", "<<upper(ue)<<"]\n";
        err << "a=RIF(["<<lower(a)<<", "<<upper(a)<<"])\n";
        err << "b=RIF(["<<lower(b)<<", "<<upper(b)<<"])\n";
        err << "c=RIF(["<<lower(c)<<", "<<upper(c)<<"])\n";
        err << "(ra,rb,rc)=("<< ra <<"," <<rb<<","<<rc<<")\n";
        err << "rs=RIF("<< lower(rs) <<"," <<upper(rs)<<")\n";
        throw err.str();
    }

    // test if it is epsilon tight
    if (upper(a-Rb-Rc)<eps and upper(b-Ra-Rc)<eps and upper(c-Ra-Rb)<eps) return ++nt;

    // test if it may be feasible
    I rs=feasible(a,b,c,Ra,Rb,Rc);
    if (upper(rs)<0) return ++nb;
    
    //one can assume that R is in [0,r] (otherwise not feasible and surely ruled out)
    rs=intersect(rs,hull(I(0),r));
    // compare emptiness and potential
    I e=E(a,b,c,Ra,Rb,Rc);
    I u=U(a,b,c,ra,rb,rc,rs);
    if (lower(e)>=upper(u)) return ++ng; // good triangles!
    if (upper(e)<lower(u)) // local inequality is not satisfied
    {
        std::ostringstream err;
        err << "Local inequality violated:\n";
        err << "E=[" << lower(e) <<", "<<upper(e)<<"], U=["<<lower(u)<<", "<<upper(u)<<"]\n";
        err << "a=RIF(["<<lower(a)<<", "<<upper(a)<<"])\n";
        err << "b=RIF(["<<lower(b)<<", "<<upper(b)<<"])\n";
        err << "c=RIF(["<<lower(c)<<", "<<upper(c)<<"])\n";
        err << "(ra,rb,rc)=("<< ra <<"," <<rb<<","<<rc<<")\n";
        err << "rs=RIF("<< lower(rs) <<"," <<upper(rs)<<")\n";
        throw err.str();
    }
    // precision does not suffice to compare emptiness and potential -> refine
    std::pair<I,I> aa=bisect(a);
    std::pair<I,I> bb=bisect(b);
    std::pair<I,I> cc=bisect(c);
    check_local_ieq(aa.first ,bb.first ,cc.first ,ra,rb,rc,depth+1);
    check_local_ieq(aa.first ,bb.first ,cc.second,ra,rb,rc,depth+1);
    check_local_ieq(aa.first ,bb.second,cc.first ,ra,rb,rc,depth+1);
    check_local_ieq(aa.first ,bb.second,cc.second,ra,rb,rc,depth+1);
    check_local_ieq(aa.second,bb.first ,cc.first ,ra,rb,rc,depth+1);
    check_local_ieq(aa.second,bb.first ,cc.second,ra,rb,rc,depth+1);
    check_local_ieq(aa.second,bb.second,cc.first ,ra,rb,rc,depth+1);
    check_local_ieq(aa.second,bb.second,cc.second,ra,rb,rc,depth+1);
    return 0;
}

